/* Copyright (C) 2009-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Valentina Akishina, Ivan Kisel, Sergey Gorbunov [committer], Igor Kulakov, Sergei Zharko, Maksym Zyzak */

/*
 *=====================================================
 *
 *  CBM Level 1 4D Reconstruction
 *
 *  Authors: V.Akishina, I.Kisel,  S.Gorbunov, I.Kulakov, M.Zyzak
 *  Documentation: V.Akishina
 *
 *  e-mail : v.akishina@gsi.de
 *
 *=====================================================
 *
 *  Finds tracks using the Cellular Automaton algorithm
 *
 */

#include "CaFramework.h"
#include "CaTimer.h"
#include "CaTrack.h"

#include <chrono>
#include <fstream>
#include <sstream>
#include <thread>

using namespace cbm::algo::ca;

using cbm::algo::ca::ECounter;
using cbm::algo::ca::ETimer;
using cbm::algo::ca::Track;


// ---------------------------------------------------------------------------------------------------------------------
//
TrackFinder::TrackFinder(ca::Framework& algo) : frAlgo(algo) {}

// ---------------------------------------------------------------------------------------------------------------------
//
TrackFinder::~TrackFinder() {}

// ---------------------------------------------------------------------------------------------------------------------
//
void TrackFinder::FindTracks()
{
  frAlgo.fRecoTracks.clear();
  frAlgo.fRecoHits.clear();
  if (frAlgo.GetInputData().GetNhits() < 1) {
    LOG(warn) << "No hits were passed to the ca::TrackFinder. Stopping the routine";
    return;
  }

  //
  // The main CaTrackFinder routine
  // It splits the input data into sub-timeslices
  // and runs the track finder over the sub-slices
  //
  frAlgo.fMonitorData.StartTimer(ETimer::Tracking);
  frAlgo.fMonitorData.StartTimer(ETimer::PrepareTimeslice);
  frAlgo.fMonitorData.IncrementCounter(ECounter::TrackingCall);
  frAlgo.fMonitorData.IncrementCounter(ECounter::RecoHit, frAlgo.fInputData.GetNhits());

  auto timerStart = std::chrono::high_resolution_clock::now();

  auto& wDataThread0 = frAlgo.fvWData[0];  // NOTE: Thread 0 must be always defined

  // ----- Reset data arrays -------------------------------------------------------------------------------------------

  wDataThread0.HitKeyFlags().reset(frAlgo.fInputData.GetNhitKeys(), 0);

  frAlgo.fHitTimeInfo.reset(frAlgo.fInputData.GetNhits());


  // TODO: move these values to Parameters namespace (S.Zharko)

  // length of sub-TS
  fscal minProtonMomentum = 0.1;

  fscal targX = frAlgo.GetParameters().GetTargetPositionX()[0];
  fscal targY = frAlgo.GetParameters().GetTargetPositionY()[0];
  fscal targZ = frAlgo.GetParameters().GetTargetPositionZ()[0];

  fStatTsStart    = std::numeric_limits<fscal>::max();  // end time of the TS
  fStatTsEnd      = 0.;                                 // end time of the TS
  fStatNhitsTotal = 0;

  // calculate possible event time for the hits (frAlgo.fHitTimeInfo array)

  const int nDataStreams = frAlgo.fInputData.GetNdataStreams();

  for (int iStream = 0; iStream < nDataStreams; ++iStream) {

    fscal maxTimeBeforeHit = std::numeric_limits<fscal>::lowest();

    int nStreamHits = frAlgo.fInputData.GetStreamNhits(iStream);
    fStatNhitsTotal += nStreamHits;

    for (int ih = 0; ih < nStreamHits; ++ih) {

      ca::HitIndex_t caHitId = frAlgo.fInputData.GetStreamStartIndex(iStream) + ih;
      const ca::Hit& h       = frAlgo.fInputData.GetHit(caHitId);

      const ca::Station<fvec>& st = frAlgo.GetParameters().GetStation(h.Station());

      fscal dx = h.X() - targX;
      fscal dy = h.Y() - targY;
      fscal dz = h.Z() - targZ;
      fscal l  = sqrt(dx * dx + dy * dy + dz * dz);

      fscal timeOfFlightMin = l * constants::phys::SpeedOfLightInv;
      fscal timeOfFlightMax =
        1.5 * l
        * sqrt(1. + constants::phys::ProtonMassD * constants::phys::ProtonMassD / minProtonMomentum / minProtonMomentum)
        * constants::phys::SpeedOfLightInvD;
      // TODO: Is it possible, that the proton mass selection affects the search of heavier particles?
      fscal dt = h.RangeT();

      CaHitTimeInfo& info = frAlgo.fHitTimeInfo[caHitId];
      info.fEventTimeMin  = h.T() - dt - timeOfFlightMax;
      info.fEventTimeMax  = h.T() + dt - timeOfFlightMin;

      if (!st.timeInfo) {
        info.fEventTimeMin = -1.e10;
        info.fEventTimeMax = 1.e10;
      }

      // NOTE: if not a MT part, use wDataThread0.IsHitKeyUsed, it will be later copied to other threads
      if (info.fEventTimeMin > 500.e6 || info.fEventTimeMax < -500.) {  // cut hits with bogus start time > 500 ms
        wDataThread0.IsHitKeyUsed(h.FrontKey()) = 1;
        wDataThread0.IsHitKeyUsed(h.BackKey())  = 1;
        LOG(error) << "CATrackFinder: skip bogus hit " << h.ToString();
        continue;
      }

      if (maxTimeBeforeHit < info.fEventTimeMax) {
        maxTimeBeforeHit = info.fEventTimeMax;
      }
      info.fMaxTimeBeforeHit = maxTimeBeforeHit;

      if (fStatTsStart > info.fEventTimeMax - fWindowMargin) {
        fStatTsStart = info.fEventTimeMax - fWindowMargin;  // 5 ns margin
      }
      if (fStatTsEnd < info.fEventTimeMin) {
        fStatTsEnd = info.fEventTimeMin;
      }
    }

    fscal minTimeAfterHit = std::numeric_limits<fscal>::max();
    // loop in the reverse order to fill CaHitTimeInfo::fMinTimeAfterHit fields

    for (int ih = nStreamHits - 1; ih >= 0; --ih) {
      ca::HitIndex_t caHitId = frAlgo.fInputData.GetStreamStartIndex(iStream) + ih;
      const ca::Hit& h       = frAlgo.fInputData.GetHit(caHitId);
      if (wDataThread0.IsHitKeyUsed(h.FrontKey()) || wDataThread0.IsHitKeyUsed(h.BackKey())) {
        continue;
      }  // the hit is skipped
      CaHitTimeInfo& info = frAlgo.fHitTimeInfo[caHitId];
      if (minTimeAfterHit > info.fEventTimeMin) {
        minTimeAfterHit = info.fEventTimeMin;
      }
      info.fMinTimeAfterHit = minTimeAfterHit;
    }

    if (0) {
      static int tmp = 0;
      if (tmp < 10000) {
        tmp++;
        LOG(warning) << "\n\n stream " << iStream << " hits " << nStreamHits << "\n\n";
        for (int ih = 0; (ih < nStreamHits) && (tmp < 10000); ++ih) {
          ca::HitIndex_t caHitId = frAlgo.fInputData.GetStreamStartIndex(iStream) + ih;
          const ca::Hit& h       = frAlgo.fInputData.GetHit(caHitId);
          if (wDataThread0.IsHitKeyUsed(h.FrontKey()) || wDataThread0.IsHitKeyUsed(h.BackKey())) {
            continue;
          }  // the hit is skipped
          CaHitTimeInfo& info = frAlgo.fHitTimeInfo[caHitId];
          if (h.Station() < 4) {
            tmp++;
            LOG(warning) << " hit sta " << h.Station() << " stream " << iStream << " time " << h.T() << " event time "
                         << info.fEventTimeMin << " .. " << info.fEventTimeMax << " max time before hit "
                         << info.fMaxTimeBeforeHit << " min time after hit " << info.fMinTimeAfterHit;
          }
        }
      }
    }
  }


  if (fStatTsEnd < fStatTsStart) {  // all hits belong to one sub-timeslice
    fStatTsEnd = fStatTsStart;
  }

  if (fStatTsEnd > fStatTsStart + 500.e6) {  // 500 ms maximal length of the TS
    fStatTsEnd = fStatTsStart + 500.e6;
  }

  LOG(debug) << "CA tracker process time slice " << fStatTsStart * 1.e-6 << " -- " << fStatTsEnd * 1.e-6
             << " [ms] with " << fStatNhitsTotal << " hits";

  int nWindows = static_cast<int>((fStatTsEnd - fStatTsStart - fWindowOverlap - fWindowMargin)
                                  / (fWindowLength - fWindowOverlap - fWindowMargin))
                 + 1;
  if (nWindows < 1) {  // Situation, when fStatTsEnd == fStatTsStart
    nWindows = 1;
  }

  int nWindowsThread = nWindows / frAlgo.GetNofThreads();
  float windowDelta  = fWindowLength - fWindowOverlap - fWindowMargin;
  //LOG(info) << "CA: estimated number of time windows: " << nWindows;


  fvWindowStartThread[0] = fStatTsStart;
  {  // Estimation of number of hits in time windows
    //Timer time;
    //time.Start();
    int nSt = frAlgo.GetParameters().GetNstationsActive();

    // Count number of hits per window and station
    double a            = fStatTsStart + fWindowOverlap + fWindowMargin;
    double b            = fWindowLength - fWindowOverlap - fWindowMargin;
    HitIndex_t nHitsTot = frAlgo.fInputData.GetNhits();
    std::vector<HitIndex_t> nHitsWindowSta(nWindows * nSt, 0);
    for (HitIndex_t iHit = 0; iHit < nHitsTot; ++iHit) {
      const auto& hit = frAlgo.fInputData.GetHit(iHit);
      //const auto& timeInfo = frAlgo.fHitTimeInfo[iHit];
      size_t iWindow = static_cast<size_t>((hit.T() - a) / b);
      if (iWindow >= static_cast<size_t>(nWindows)) {
        LOG(error) << "ca: Hit out of range: iHit = " << iHit << ", time = " << hit.T() * 1.e-6
                   << " ms, window = " << iWindow;
        continue;
      }

      ++nHitsWindowSta[hit.Station() + iWindow * nSt];
    }

    // Remove hits from the "monster events"
    if (ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
      auto maxNofHitsSta = static_cast<HitIndex_t>(50 * fWindowLength / 1.e3);
      for (auto& content : nHitsWindowSta) {
        if (content > maxNofHitsSta) {
          content = 0;
        }
      }
    }

    // Integrate number of hits per window
    std::vector<HitIndex_t> nHitsWindow;
    nHitsWindow.reserve(nWindows);
    auto windowStaIt          = nHitsWindowSta.begin();
    HitIndex_t nHitsCollected = 0;
    while (windowStaIt != nHitsWindowSta.end()) {
      nHitsCollected = nHitsWindow.emplace_back(std::accumulate(windowStaIt, windowStaIt + nSt, nHitsCollected));
      windowStaIt    = windowStaIt + nSt;
    }

    // Get time range for threads
    HitIndex_t nHitsPerThread = nHitsCollected / frAlgo.GetNofThreads();
    auto windowIt             = nHitsWindow.begin();
    size_t iWbegin            = 0;
    for (int iTh = 1; iTh < frAlgo.GetNofThreads(); ++iTh) {
      windowIt                   = std::lower_bound(windowIt, nHitsWindow.end(), iTh * nHitsPerThread);
      iWbegin                    = std::distance(nHitsWindow.begin(), windowIt) + 1;
      fvWindowStartThread[iTh]   = fStatTsStart + iWbegin * windowDelta;
      fvWindowEndThread[iTh - 1] = fvWindowStartThread[iTh];
    }
    //time.Stop();
    //LOG(info) << "Thread boarders estimation time: " << time.GetTotalMs() << " ms";
    LOG(info) << "Fraction of hits from monster events: " << static_cast<double>(nHitsTot - nHitsCollected) / nHitsTot;
  }


  // cut data into sub-timeslices and process them one by one
  //for (int iThread = 0; iThread < frAlgo.GetNofThreads(); ++iThread) {
  //  fvWindowStartThread[iThread] = fStatTsStart + iThread * nWindowsThread * windowDelta;
  //  fvWindowEndThread[iThread] =
  //    fvWindowStartThread[iThread] + nWindowsThread * windowDelta + fWindowOverlap + fWindowMargin;
  //}
  fvWindowEndThread[frAlgo.GetNofThreads() - 1] = fStatTsEnd;

  for (int iThread = 0; iThread < frAlgo.GetNofThreads(); ++iThread) {
    double start = fvWindowStartThread[iThread] * 1.e-6;
    double end   = fvWindowEndThread[iThread] * 1.e-6;
    LOG(trace) << "Thread: " << iThread << " from " << start << " ms  to " << end << " ms (delta = " << end - start
               << " ms)";
  }

  frAlgo.fMonitorData.StopTimer(ETimer::PrepareTimeslice);
  // Save tracks
  if (frAlgo.GetNofThreads() == 1) {
    this->FindTracksThread(0);
    frAlgo.fMonitorData.StartTimer(ETimer::StoreTracksFinal);
    frAlgo.fRecoTracks = std::move(fvRecoTracks[0]);
    frAlgo.fRecoHits   = std::move(fvRecoHitIndices[0]);
    frAlgo.fMonitorData.StopTimer(ETimer::StoreTracksFinal);
  }
  else {
    std::vector<std::thread> vThreadList;
    vThreadList.reserve(frAlgo.GetNofThreads());
    for (int iTh = 0; iTh < frAlgo.GetNofThreads(); ++iTh) {
      vThreadList.emplace_back(&TrackFinder::FindTracksThread, this, iTh);
    }
    for (auto& th : vThreadList) {
      if (th.joinable()) {
        th.join();
      }
    }
    frAlgo.fMonitorData.StartTimer(ETimer::StoreTracksFinal);
    auto Operation  = [](size_t acc, const auto& v) { return acc + v.size(); };
    int nRecoTracks = std::accumulate(fvRecoTracks.begin(), fvRecoTracks.end(), 0, Operation);
    int nRecoHits   = std::accumulate(fvRecoHitIndices.begin(), fvRecoHitIndices.end(), 0, Operation);
    frAlgo.fRecoTracks.reserve(nRecoTracks);
    frAlgo.fRecoHits.reserve(nRecoHits);
    for (int iTh = 0; iTh < frAlgo.GetNofThreads(); ++iTh) {
      frAlgo.fRecoTracks.insert(frAlgo.fRecoTracks.end(), fvRecoTracks[iTh].begin(), fvRecoTracks[iTh].end());
      frAlgo.fRecoHits.insert(frAlgo.fRecoHits.end(), fvRecoHitIndices[iTh].begin(), fvRecoHitIndices[iTh].end());
    }
    frAlgo.fMonitorData.StopTimer(ETimer::StoreTracksFinal);
  }

  frAlgo.fMonitorData.IncrementCounter(ECounter::RecoTrack, frAlgo.fRecoTracks.size());
  frAlgo.fMonitorData.IncrementCounter(ECounter::RecoHitUsed, frAlgo.fRecoHits.size());

  auto timerEnd      = std::chrono::high_resolution_clock::now();
  frAlgo.fCaRecoTime = (double) (std::chrono::duration<double>(timerEnd - timerStart).count());

  // Add thread monitors to the main monitor
  for (auto& monitor : frAlgo.fvMonitorDataThread) {
    frAlgo.fMonitorData.AddMonitorData(monitor, true);
    //frAlgo.fMonitorData.AddMonitorData(monitor);
    monitor.Reset();
  }

  int statNhitsProcessedTotal = 0;
  int statNwindowsTotal       = 0;
  for (int iThread = 0; iThread < frAlgo.GetNofThreads(); ++iThread) {
    statNhitsProcessedTotal += fvStatNhitsProcessed[iThread];
    statNwindowsTotal += fvStatNwindows[iThread];
  }

  // Filling TS headear
  auto& tsHeader   = frAlgo.fTsHeader;
  tsHeader.Start() = fStatTsStart;
  tsHeader.End()   = fStatTsEnd;

  frAlgo.fMonitorData.StopTimer(ETimer::Tracking);

  LOG(debug) << "CA tracker: time slice finished. Reconstructed " << frAlgo.fRecoTracks.size() << " tracks with "
             << frAlgo.fRecoHits.size() << " hits. Processed " << statNhitsProcessedTotal << " hits in "
             << statNwindowsTotal << " time windows. Reco time " << frAlgo.fCaRecoTime / 1.e9 << " s";
}

// ---------------------------------------------------------------------------------------------------------------------
//
void TrackFinder::FindTracksThread(int iThread)
{
  //std::stringstream filename;
  //filename << "./dbg_caTrackFinder::FindTracksThread_" << iThread << ".txt";
  //std::ofstream out(filename.str());
  auto& monitor = frAlgo.fvMonitorDataThread[iThread];
  Timer timer;
  timer.Start();
  monitor.StartTimer(ETimer::TrackingThread);
  monitor.StartTimer(ETimer::PrepareThread);

  // Init vectors
  auto& wData = frAlgo.fvWData[iThread];
  {
    int nStations          = frAlgo.GetParameters().GetNstationsActive();
    size_t nHitsTot        = frAlgo.fInputData.GetNhits();
    size_t nHitKeys        = frAlgo.fInputData.GetNhitKeys();
    size_t nHitsExpected   = 2 * nHitsTot;
    size_t nTracksExpected = 2 * nHitsTot / nStations;
    size_t nThreads        = frAlgo.GetNofThreads();
    fvRecoTracks[iThread].clear();
    fvRecoTracks[iThread].reserve(nTracksExpected / nThreads);
    fvRecoHitIndices[iThread].clear();
    fvRecoHitIndices[iThread].reserve(nHitsExpected / nThreads);
    frAlgo.fvTrackFinderWindow[iThread].InitTimeslice();
    if (iThread != 0) {
      auto& wData0 = frAlgo.fvWData[0];
      // slow ...
      wData.HitKeyFlags().clear();
      wData.HitKeyFlags().reserve(nHitKeys);
      for (const auto& flag : wData0.HitKeyFlags()) {
        wData.HitKeyFlags().emplace_back(flag);
      }
    }
    for (int iS = 0; iS < nStations; ++iS) {
      wData.TsHitIndices(iS).clear();
      wData.TsHitIndices(iS).reserve(nHitsTot);
    }
  }

  const int nDataStreams    = frAlgo.fInputData.GetNdataStreams();
  bool areUntouchedDataLeft = true;  // is the whole TS processed
  std::vector<HitIndex_t> sliceFirstHit(nDataStreams, 0);
  std::vector<HitIndex_t> sliceLastHit(nDataStreams, 0);

  // Define first hit, skip all the hits, which are before the first window
  for (int iStream = 0; iStream < nDataStreams; ++iStream) {
    sliceFirstHit[iStream] = frAlgo.fInputData.GetStreamStartIndex(iStream);
    sliceLastHit[iStream]  = frAlgo.fInputData.GetStreamStopIndex(iStream);
    for (HitIndex_t caHitId = sliceFirstHit[iStream]; caHitId < sliceLastHit[iStream]; ++caHitId) {
      CaHitTimeInfo& info = frAlgo.fHitTimeInfo[caHitId];
      const ca::Hit& h    = frAlgo.fInputData.GetHit(caHitId);
      if (wData.IsHitKeyUsed(h.FrontKey()) || wData.IsHitKeyUsed(h.BackKey())) {
        continue;
      }
      if (info.fMaxTimeBeforeHit < fvWindowStartThread[iThread]) {
        sliceFirstHit[iStream] = caHitId + 1;
      }
      if (info.fMinTimeAfterHit > fvWindowEndThread[iThread]) {
        sliceLastHit[iStream] = caHitId;
      }
    }
  }

  fvStatNwindows[iThread]       = 0;
  fvStatNhitsProcessed[iThread] = 0;

  int statLastLogTimeChunk = -1;

  monitor.StopTimer(ETimer::PrepareThread);

  while (areUntouchedDataLeft) {
    frAlgo.fvMonitorDataThread[iThread].IncrementCounter(ECounter::SubTS);
    // select the sub-slice hits
    for (int iS = 0; iS < frAlgo.GetParameters().GetNstationsActive(); ++iS) {
      frAlgo.fvWData[iThread].TsHitIndices(iS).clear();
    }

    areUntouchedDataLeft = false;

    // TODO: SG: skip empty regions and start the subslice with the earliest hit

    fvStatNwindows[iThread]++;
    //out << fvStatNwindows[iThread] << ' ';

    int statNwindowHits = 0;

    frAlgo.fvMonitorDataThread[iThread].StartTimer(ETimer::PrepareWindow);
    for (int iStream = 0; iStream < nDataStreams; ++iStream) {
      for (ca::HitIndex_t caHitId = sliceFirstHit[iStream]; caHitId < sliceLastHit[iStream]; ++caHitId) {
        const CaHitTimeInfo& info = frAlgo.fHitTimeInfo[caHitId];
        const ca::Hit& h          = frAlgo.fInputData.GetHit(caHitId);
        if (wData.IsHitKeyUsed(h.FrontKey()) || wData.IsHitKeyUsed(h.BackKey())) {
          // the hit is already reconstructed
          continue;
        }
        if (info.fEventTimeMin
            > fvWindowStartThread[iThread] + fWindowLength) {  // the hit is too late for the sub slice
          areUntouchedDataLeft = true;
          if (info.fMinTimeAfterHit > fvWindowStartThread[iThread] + fWindowLength) {
            // this hit and all later hits are out of the sub-slice
            break;
          }
        }
        else {
          if (fvWindowStartThread[iThread] <= info.fEventTimeMax) {  // the hit belongs to the sub-slice
            frAlgo.fvWData[iThread].TsHitIndices(h.Station()).push_back(caHitId);
            statNwindowHits++;
            if (info.fMaxTimeBeforeHit < fvWindowStartThread[iThread] + fWindowLength - fWindowOverlap) {
              // this hit and all hits before are before the overlap
              sliceFirstHit[iStream] = caHitId + 1;
            }
          }
        }  // else the hit has been alread processed in previous sub-slices
      }
    }

    //out << statNwindowHits << ' ';
    //if (statNwindowHits == 0) {  // Empty window
    //  frAlgo.fvMonitorDataThread[iThread].StopTimer(ETimer::PrepareWindow);
    //  out << 0 << ' ' << 0 << ' ' << 0 << '\n';
    //  continue;
    //}

    if (ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
      // cut at 50 hits per station per 1 us.
      int maxStationHits = (int) (50 * fWindowLength / 1.e3);
      for (int ista = 0; ista < frAlgo.GetParameters().GetNstationsActive(); ++ista) {
        int nHitsSta = static_cast<int>(frAlgo.fvWData[iThread].TsHitIndices(ista).size());
        if (nHitsSta > maxStationHits) {
          statNwindowHits -= nHitsSta;
          frAlgo.fvWData[iThread].TsHitIndices(ista).clear();
        }
      }
    }

    fvStatNhitsProcessed[iThread] += statNwindowHits;

    // print the LOG for every 10 ms of data processed
    if constexpr (0) {
      int currentChunk = (int) ((fvWindowStartThread[iThread] - fStatTsStart) / 10.e6);
      if (!areUntouchedDataLeft || currentChunk > statLastLogTimeChunk) {
        statLastLogTimeChunk = currentChunk;
        double dataRead =
          100. * (fvWindowStartThread[iThread] + fWindowLength - fStatTsStart) / (fStatTsEnd - fStatTsStart);
        if (dataRead > 100.) {
          dataRead = 100.;
        }
        LOG(info) << "CA tracker process sliding window N " << fvStatNwindows[iThread] << ": time "
                  << fvWindowStartThread[iThread] / 1.e6 << " ms + " << fWindowLength / 1.e3 << " us) with "
                  << statNwindowHits << " hits. "
                  << " Processing " << dataRead << " % of the TS time and "
                  << 100. * fvStatNhitsProcessed[iThread] / fStatNhitsTotal << " % of TS hits."
                  << " Already reconstructed " << fvRecoTracks[iThread].size() << " tracks on thread #" << iThread;
      }
    }


    //out << statNwindowHits << ' ';

    frAlgo.fvMonitorDataThread[iThread].StopTimer(ETimer::PrepareWindow);

    //Timer trackingInWindow;  //DBG
    //trackingInWindow.Start();
    frAlgo.fvMonitorDataThread[iThread].StartTimer(ETimer::TrackingWindow);
    frAlgo.fvTrackFinderWindow[iThread].CaTrackFinderSlice();
    frAlgo.fvMonitorDataThread[iThread].StopTimer(ETimer::TrackingWindow);
    //trackingInWindow.Stop();
    //out << trackingInWindow.GetTotalMs() << ' ';

    // save reconstructed tracks with no hits in the overlap region
    //if (fvWindowStartThread[iThread] > 13.23e6 && fvWindowStartThread[iThread] < 13.26e6) {
    fvWindowStartThread[iThread] += fWindowLength - fWindowOverlap;
    // we should add hits from reconstructed but not stored tracks to the new sub-timeslice
    // we do it in a simple way by extending the tsStartNew
    // TODO: only add those hits from the region before tsStartNew that belong to the not stored tracks
    //out << frAlgo.fvWData[iThread].RecoHitIndices().size() << ' ';
    //out << frAlgo.fvWData[iThread].RecoTracks().size() << '\n';

    frAlgo.fvMonitorDataThread[iThread].StartTimer(ETimer::StoreTracksWindow);
    int trackFirstHit = 0;
    for (const auto& track : frAlgo.fvWData[iThread].RecoTracks()) {
      bool isTrackCompletelyInOverlap = true;
      for (int ih = 0; ih < track.fNofHits; ih++) {
        int caHitId         = frAlgo.fvWData[iThread].RecoHitIndex(trackFirstHit + ih);
        CaHitTimeInfo& info = frAlgo.fHitTimeInfo[caHitId];
        if (info.fEventTimeMax < fvWindowStartThread[iThread]) {  // this hit is before the overlap
          isTrackCompletelyInOverlap = false;
          break;
        }
      }

      if (!areUntouchedDataLeft) {  // don't reject tracks in the overlap when no more data are left
        isTrackCompletelyInOverlap = 0;
      }

      if (isTrackCompletelyInOverlap) {
        //
        // Don't save tracks from the overlap region, since they might have additional hits in the next subslice
        //

        // release the track hits
        for (int i = 0; i < track.fNofHits; i++) {
          int caHitId                      = frAlgo.fvWData[iThread].RecoHitIndex(trackFirstHit + i);
          const auto& h                    = frAlgo.fInputData.GetHit(caHitId);
          wData.IsHitKeyUsed(h.FrontKey()) = 0;
          wData.IsHitKeyUsed(h.BackKey())  = 0;
        }
      }
      else {  // save the track
        fvRecoTracks[iThread].push_back(track);
        // mark the track hits as used
        for (int i = 0; i < track.fNofHits; i++) {
          int caHitId                      = frAlgo.fvWData[iThread].RecoHitIndex(trackFirstHit + i);
          const auto& h                    = frAlgo.fInputData.GetHit(caHitId);
          wData.IsHitKeyUsed(h.FrontKey()) = 1;
          wData.IsHitKeyUsed(h.BackKey())  = 1;
          fvRecoHitIndices[iThread].push_back(caHitId);
        }
      }
      trackFirstHit += track.fNofHits;
    }  // sub-timeslice tracks
    frAlgo.fvMonitorDataThread[iThread].StopTimer(ETimer::StoreTracksWindow);

    if (fvWindowStartThread[iThread] > fvWindowEndThread[iThread]) {
      break;
    }
    else {
      fvWindowStartThread[iThread] -= fWindowMargin;  // do 5 ns margin
    }
  }
  frAlgo.fvMonitorDataThread[iThread].StopTimer(ETimer::TrackingThread);
  //timer.Stop();
  //LOG(info) << "CA: finishing tracking on thread " << iThread << " (time: " << timer.GetTotalMs() << " ms, "
  //          << "hits processed: " << fvStatNhitsProcessed[iThread] << ", "
  //          << "hits used: " << fvRecoHitIndices[iThread].size() << ')';
}

// ---------------------------------------------------------------------------------------------------------------------
//
void TrackFinder::Init()
{
  assert(frAlgo.GetNofThreads() > 0);

  fvWindowStartThread.clear();
  fvWindowStartThread.resize(frAlgo.GetNofThreads());
  fvWindowEndThread.clear();
  fvWindowEndThread.resize(frAlgo.GetNofThreads());
  fvStatNwindows.clear();
  fvStatNwindows.resize(frAlgo.GetNofThreads());
  fvStatNhitsProcessed.clear();
  fvStatNhitsProcessed.resize(frAlgo.GetNofThreads());
  fvRecoTracks.clear();
  fvRecoTracks.resize(frAlgo.GetNofThreads());
  fvRecoHitIndices.clear();
  fvRecoHitIndices.resize(frAlgo.GetNofThreads());

  for (int iThread = 0; iThread < frAlgo.GetNofThreads(); ++iThread) {
    fvRecoTracks[iThread].SetName(std::string("TrackFinder::fvRecoTracks_") + std::to_string(iThread));
    fvRecoHitIndices[iThread].SetName(std::string("TrackFinder::fvRecoHitIndices_") + std::to_string(iThread));
  }

  fWindowLength = (ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) ? 500 : 10000;
}
