/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */

/// \file CaTripletConstructor.h
/// \brief Triplet constructor for the CA tracker
/// \author Sergey Gorbunov

#pragma once  // include this header only once per compilation unit

#include "CaField.h"
#include "CaFramework.h"
#include "CaGridEntry.h"
#include "CaStation.h"
#include "CaTrackFit.h"
#include "CaTrackParam.h"
#include "CaTriplet.h"
#include "CaVector.h"
#include "CaWindowData.h"

namespace cbm::algo::ca
{

  namespace
  {
    using namespace cbm::algo;  // to keep ca:: prefices in the code
  }

  /// Construction of triplets for the CA tracker
  ///
  class TripletConstructor {
   public:
    ///                             ------  Constructors and destructor ------

    /// Constructor
    /// \param nThreads  Number of threads for multi-threaded mode
    TripletConstructor(ca::Framework& algo, ca::WindowData& wData);

    /// Copy constructor
    TripletConstructor(const TripletConstructor&) = delete;

    /// Move constructor
    TripletConstructor(TripletConstructor&&) = delete;

    /// Copy assignment operator
    TripletConstructor& operator=(const TripletConstructor&) = delete;

    /// Move assignment operator
    TripletConstructor& operator=(TripletConstructor&&) = delete;

    /// Destructor
    ~TripletConstructor() = default;


    ///                             ------  FUNCTIONAL PART ------

    void InitStations(int istal, int istam, int istar);

    void CreateTripletsForHit(int istal, int istam, int istar, ca::HitIndex_t ihl);

    const Vector<ca::Triplet>& GetTriplets() const { return fTriplets; }

    /// Find the doublets. Reformat data in the portion of doublets.
    void FindDoublets();
    void FitDoublets();

    /// Find triplets on station
    void FindTriplets();

    /// Add the middle hits to parameters estimation. Propagate to right station.
    /// Find the triplets (right hit). Reformat data in the portion of triplets.
    void FindRightHit();

    /// Fit Triplets
    void FitTriplets();

    /// Select triplets. Save them into vTriplets.
    void StoreTriplets();

    void CollectHits(const TrackParamV& Tr, const int iSta, const double chi2Cut, const int iMC,
                     Vector<ca::HitIndex_t>& collectedHits, int maxNhits);

   private:
    /// left station
    const ca::Station<fvec>& staL() { return *fStaL; }
    /// middle station
    const ca::Station<fvec>& staM() { return *fStaM; }
    /// right station
    const ca::Station<fvec>& staR() { return *fStaR; }

   private:
    bool fIsInitialized{false};  ///< is initialized;
    ca::Framework& frAlgo;       ///< reference to ca::Framework object

    int fIstaL{-1};  ///< left station index
    int fIstaM{-1};  ///< middle station index
    int fIstaR{-1};  ///< right station index

    const ca::Station<fvec>* fStaL{nullptr};  ///< left station
    const ca::Station<fvec>* fStaM{nullptr};  ///< mid station
    const ca::Station<fvec>* fStaR{nullptr};  ///< right station

    const ca::Station<fvec>*
      fFld0Sta[2];  // two stations for approximating the field between the target and the left hit
    const ca::Station<fvec>*
      fFld1Sta[3];  // three stations for approximating the field between the left and the right hit

    ca::HitIndex_t fIhitL;  ///< index of the left hit in fAlgo->fWindowHits
    TrackParamV fTrL;
    ca::FieldRegion<fvec> fFldL;

    Vector<ca::HitIndex_t> fHitsM_2{"TripletConstructor::fHitsM_2"};
    Vector<TrackParamV> fTracks_2{"TripletConstructor::fTracks_2"};

    Vector<TrackParamV> fTracks_3{"TripletConstructor::fTracks_3"};
    Vector<ca::HitIndex_t> fHitsM_3{"TripletConstructor::fHitsM_3"};
    Vector<ca::HitIndex_t> fHitsR_3{"TripletConstructor::fHitsR_3"};

    Vector<ca::Triplet> fTriplets{"TripletConstructor::fTriplets"};

    ca::TrackFit fFit;
    WindowData& frWData;

   private:
    static constexpr bool fDebugDublets     = false;  // print debug info for dublets
    static constexpr bool fDebugTriplets    = false;  // print debug info for triplets
    static constexpr bool fDebugCollectHits = false;  // print debug info for CollectHits

  } _fvecalignment;

}  // namespace cbm::algo::ca
