/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   StsDigiQa.h
/// \date   03.03.2024
/// \brief  QA module for STS raw digis (source)
/// \author Sergei Zharko <s.zharko@gsi.de>

#include "StsDigiQa.h"

#include "CbmStsAddress.h"

#include <fmt/format.h>

using cbm::algo::sts::DigiQa;
using fmt::format;

// ---------------------------------------------------------------------------------------------------------------------
//
void DigiQa::Init()
{
  using cbm::algo::qa::CanvasConfig;
  using cbm::algo::qa::Data;
  using cbm::algo::qa::H1D;
  using cbm::algo::qa::H2D;
  using cbm::algo::qa::PadConfig;

  if (!fpSender.get()) {
    return;
  }

  int nModules = fpReadoutSetup->modules.size();

  // Histograms per address
  {
    fvphAddressChannel.resize(nModules);
    fvphAddressCharge.resize(nModules);
    fvphAddressChannelCharge.resize(nModules);
    for (int iM = 0; iM < nModules; ++iM) {
      int32_t address       = fpReadoutSetup->modules.at(iM).address;
      fmAddressMap[address] = iM;

      auto cName = format("sts_digi/sts_digi_vs_channel_charge_addr{:#10x}", address);
      auto cTitl = format("STS digis per channel and charge for module {:#10x}", address);
      auto canv  = CanvasConfig(cName, cTitl, 3, 1);
      {
        auto pad               = PadConfig();
        auto name              = format("sts_digi_addr{}_channel", iM);
        auto titl              = format("Number of digis per channel for address {:#10x};channel;N_{{digis}}", address);
        fvphAddressChannel[iM] = fQaData.MakeObj<H1D>(name, titl, 2048, -0.5, 2047.5);
        pad.RegisterHistogram(fvphAddressChannel[iM], "hist");
        canv.AddPadConfig(pad);
      }
      {
        auto pad              = PadConfig();
        auto name             = format("sts_digi_addr{}_charge", iM);
        auto titl             = format("STS digi charge for address {:#10x};charge [ADS units];N_{{digis}}", address);
        fvphAddressCharge[iM] = fQaData.MakeObj<H1D>(name, titl, 32, -0.5, 31.5);
        pad.RegisterHistogram(fvphAddressCharge[iM], "hist");
        canv.AddPadConfig(pad);
      }
      {
        auto pad  = PadConfig();
        auto name = format("sts_digi_addr{}_channel_charge", iM);
        auto titl = format("STS digi charge for address {:#10x};charge [ADS units];channel", address);
        fvphAddressChannelCharge[iM] = fQaData.MakeObj<H2D>(name, titl, 32, -0.5, 31.5, 2048, -0.5, 2047.5);
        pad.RegisterHistogram(fvphAddressChannelCharge[iM], "colz");
        canv.AddPadConfig(pad);
      }
      fQaData.AddCanvasConfig(canv);
    }
  }

  fQaData.Init(fpSender);
}

// ---------------------------------------------------------------------------------------------------------------------
//
void DigiQa::Exec()
{
  if (!fpSender.get()) {
    return;
  }

  // Loop over STS digis and fill histograms
  for (const auto& digi : (*fpvDigis)) {
    int32_t address = digi.GetAddress();
    double channel  = digi.GetChannel();
    double charge   = digi.GetCharge();

    //  Ensure, that the address is defined
    auto itHistID = fmAddressMap.find(address);
    if (itHistID == fmAddressMap.end()) {
      L_(error) << format("std::DigiQa: found address {:x}, which is not defined in the ReadoutSetup config", address);
      continue;
    }

    int iM = itHistID->second;
    fvphAddressChannel[iM]->Fill(channel);
    fvphAddressCharge[iM]->Fill(charge);
    fvphAddressChannelCharge[iM]->Fill(charge, channel);
  }


  fQaData.Send(fpSender);
}
