/* Copyright (C) 2024 Hulubei National Institute of Physics and Nuclear Engineering - Horia Hulubei, Bucharest
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Alexandru Bercuci [committer] */

// CBM headers
#include "CbmRecoQaTask.h"

#include "CbmGlobalTrack.h"
#include "CbmTimeSlice.h"
#include "CbmTofAddress.h"
#include "CbmTrdAddress.h"
// FAIR headers
#include "FairRootManager.h"
#include "FairSink.h"
// ROOT headers
#include "TClonesArray.h"
#include "TH2D.h"

using namespace std;
//_____________________________________________________________________
CbmRecoQaTask::CbmRecoQaTask() : FairTask("RecoQA", 0) {}

//_____________________________________________________________________
InitStatus CbmRecoQaTask::Init()
{
  fFitter.Init();
  fFitter.SetSkipUnmeasuredCoordinates(true);

  //Get ROOT Manager
  FairRootManager* ioman = FairRootManager::Instance();

  if (!ioman) {
    LOG(fatal) << GetName() << "::Init :: RootManager not instantiated!";
    return kERROR;
  }

  fTracks = static_cast<TClonesArray*>(ioman->GetObject("GlobalTrack"));
  if (!fTracks) {
    LOG(fatal) << GetName() << "::Init: Global track array not found!";
    return kERROR;
  }

  fTimeSlice = static_cast<CbmTimeSlice*>(ioman->GetObject("TimeSlice."));
  if (!fTimeSlice) {
    LOG(warn) << GetName() << "::Init: No time slice object. Checking for events";
    fEvents = static_cast<TClonesArray*>(ioman->GetObject("CbmEvent"));
    if (!fEvents) LOG(warn) << GetName() << "::Init: No event found. Some results will be missing.";
  }

  fTrackMatches = static_cast<TClonesArray*>(ioman->GetObject("GlobalTrackMatch"));
  if (!fTrackMatches) {
    LOG(warn) << GetName() << "::Init: MC info for Global track not available !";
  }

  LOG(info) << GetName() << "::Init: Registered projections " << fProjView.size();

  int itrkLy(0), iprjLy(0);
  TFolder *modDir(nullptr), *smDir(nullptr);
  fOutFolder.Clear();
  switch (fSetupClass) {
    case kMcbm22:
      LOG(info) << GetName() << "::Init: Setup config for \"mCBM 2022\".";
      // STS U0 - station 0
      modDir = fOutFolder.AddFolder("STS", "STS reziduals to CA");
      fModView[itrkLy].push_back(ModuleView());
      fModView[itrkLy][0].hdXx.first = 10;
      fModView[itrkLy][0].hdXx.second =
        new TH2D(Form("hxx%d", itrkLy), Form("STS U0 x-dx; X_{STS U0} (cm); #Delta X (mm)"), 160, -8, 8, 400, -2, 2);
      modDir->Add(fModView[itrkLy][0].hdXx.second);
      fModView[itrkLy][0].hdYy.first  = 10;
      fModView[itrkLy][0].hdYy.second = new TH2D(
        Form("hyy%d", itrkLy), Form("STS U0 y-dy; Y_{STS U0} (cm); #Delta Y (mm)"), 160, -8, 8, 350, -3.5, 3.5);
      modDir->Add(fModView[itrkLy][0].hdYy.second);
      fModView[itrkLy][0].hdT.first = 1;
      fModView[itrkLy][0].hdT.second =
        new TH2D(Form("ht%d", itrkLy), Form("STS U0 x-dt; X_{STS U0} (cm); #Delta T (ns)"), 160, -8, 8, 15, -60, 60);
      modDir->Add(fModView[itrkLy][0].hdT.second);
      fModView[itrkLy][0].hpullXphi =
        new TH2D(Form("hpullx%d", itrkLy), Form("STS U0 pullx-phi; dx/dz_{trk}; pull X"), 160, -0.8, 0.8, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullXphi);
      fModView[itrkLy][0].hpullYtht =
        new TH2D(Form("hpully%d", itrkLy), Form("STS U0 pully-tht; dy/dz_{trk}; pull Y"), 160, -0.8, 0.8, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullYtht);
      itrkLy++;
      // STS U1 - station 1
      fModView[itrkLy].push_back(ModuleView());
      fModView[itrkLy][0].hdXx.first = 10;
      fModView[itrkLy][0].hdXx.second =
        new TH2D(Form("hxx%d", itrkLy), Form("STS U1 x-dx; X_{STS U1} (cm); #Delta X (mm)"), 200, -10, 10, 400, -2, 2);
      modDir->Add(fModView[itrkLy][0].hdXx.second);
      fModView[itrkLy][0].hdYy.first  = 10;
      fModView[itrkLy][0].hdYy.second = new TH2D(
        Form("hyy%d", itrkLy), Form("STS U1 y-dy; Y_{STS U1} (cm); #Delta Y (mm)"), 200, -10, 10, 350, -3.5, 3.5);
      modDir->Add(fModView[itrkLy][0].hdYy.second);
      fModView[itrkLy][0].hdT.first = 1;
      fModView[itrkLy][0].hdT.second =
        new TH2D(Form("ht%d", itrkLy), Form("STS U1 x-dt; X_{STS U1} (cm); #Delta T (ns)"), 200, -10, 10, 15, -60, 60);
      modDir->Add(fModView[itrkLy][0].hdT.second);
      fModView[itrkLy][0].hpullXphi =
        new TH2D(Form("hpullx%d", itrkLy), Form("STS U1 pullx-phi; dx/dz_{trk}; pull X"), 160, -0.8, 0.8, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullXphi);
      fModView[itrkLy][0].hpullYtht =
        new TH2D(Form("hpully%d", itrkLy), Form("STS U1 pully-tht; dy/dz_{trk}; pull Y"), 160, -0.8, 0.8, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullYtht);
      itrkLy++;

      // ===============================================================================
      modDir = fOutFolder.AddFolder("TRD", "TRD reziduals to CA");
      // TRD 2D - station 2
      fModView[itrkLy].push_back(ModuleView());
      fModView[itrkLy][0].hdXx.first = 10;
      fModView[itrkLy][0].hdXx.second =
        new TH2D(Form("hxx%d", itrkLy), Form("TRD2D x-dx; X_{TRD2D} (cm); #Delta X (mm)"), 500, -25, 25, 240, -6, 6);
      modDir->Add(fModView[itrkLy][0].hdXx.second);
      fModView[itrkLy][0].hdYy.first = 1;
      fModView[itrkLy][0].hdYy.second =
        new TH2D(Form("hyy%d", itrkLy), Form("TRD2D y-dy; Y_{TRD2D} (cm); #Delta Y (cm)"), 300, -30, 30, 400, -2, 2);
      modDir->Add(fModView[itrkLy][0].hdYy.second);
      fModView[itrkLy][0].hdT.first = 1;
      fModView[itrkLy][0].hdT.second =
        new TH2D(Form("ht%d", itrkLy), Form("TRD2D x-dt; X_{TRD2D} (cm); #Delta T (ns)"), 500, -25, 25, 25, -100, 100);
      modDir->Add(fModView[itrkLy][0].hdT.second);
      fModView[itrkLy][0].hpullXphi =
        new TH2D(Form("hpullx%d", itrkLy), Form("TRD2D pullx-phi; dx/dz_{trk}; pull X"), 160, -0.8, 0.8, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullXphi);
      fModView[itrkLy][0].hpullYtht =
        new TH2D(Form("hpully%d", itrkLy), Form("TRD2D pully-tht; dy/dz_{trk}; pull Y"), 160, -0.8, 0.8, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullYtht);
      itrkLy++;
      // TRD 1Dx - station 3
      fModView[itrkLy].push_back(ModuleView());
      fModView[itrkLy][0].hdXx.first  = 10;
      fModView[itrkLy][0].hdXx.second = new TH2D(
        Form("hxx%d", itrkLy), Form("TRD1Dx x-dx; X_{TRD1Dx} (cm); #Delta X (mm)"), 450, -45, 45, 200, -10, 10);
      modDir->Add(fModView[itrkLy][0].hdXx.second);
      fModView[itrkLy][0].hdYy.first = 1.;
      fModView[itrkLy][0].hdYy.second =
        new TH2D(Form("hyy%d", itrkLy), Form("TRD1Dx y-dy; Y_{TRD1Dx} (cm); #Delta Y (cm)"), 12, -45, 45, 400, -20, 20);
      modDir->Add(fModView[itrkLy][0].hdYy.second);
      fModView[itrkLy][0].hdT.first  = 1;
      fModView[itrkLy][0].hdT.second = new TH2D(
        Form("ht%d", itrkLy), Form("TRD1Dx x-dt; X_{TRD1Dx} (cm); #Delta T (ns)"), 450, -45, 45, 25, -100, 100);
      modDir->Add(fModView[itrkLy][0].hdT.second);
      fModView[itrkLy][0].hpullXphi =
        new TH2D(Form("hpullx%d", itrkLy), Form("TRD1Dx pullx-phi; dx/dz_{trk}; pull X"), 100, -0.5, 0.5, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullXphi);
      fModView[itrkLy][0].hpullYtht =
        new TH2D(Form("hpully%d", itrkLy), Form("TRD1Dx pully-tht; dy/dz_{trk}; pull Y"), 100, -0.5, 0.5, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullYtht);
      itrkLy++;
      // TRD 1Dy - station 4
      fModView[itrkLy].push_back(ModuleView());
      fModView[itrkLy][0].hdXx.first = 1;
      fModView[itrkLy][0].hdXx.second =
        new TH2D(Form("hxx%d", itrkLy), Form("TRD1Dy x-dx; X_{TRD1Dy} (cm); #Delta X (cm)"), 12, -45, 45, 400, -20, 20);
      modDir->Add(fModView[itrkLy][0].hdXx.second);
      fModView[itrkLy][0].hdYy.first  = 10.;
      fModView[itrkLy][0].hdYy.second = new TH2D(
        Form("hyy%d", itrkLy), Form("TRD1Dy y-dy; Y_{TRD1Dy} (cm); #Delta Y (mm)"), 500, -50, 50, 400, -20, 20);
      modDir->Add(fModView[itrkLy][0].hdYy.second);
      fModView[itrkLy][0].hdT.first  = 1;
      fModView[itrkLy][0].hdT.second = new TH2D(
        Form("ht%d", itrkLy), Form("TRD1Dy y-dt; Y_{TRD1Dy} (cm); #Delta T (ns)"), 500, -50, 50, 25, -100, 100);
      modDir->Add(fModView[itrkLy][0].hdT.second);
      fModView[itrkLy][0].hpullXphi =
        new TH2D(Form("hpullx%d", itrkLy), Form("TRD1Dy pullx-phi; dx/dz_{trk}; pull X"), 100, -0.5, 0.5, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullXphi);
      fModView[itrkLy][0].hpullYtht =
        new TH2D(Form("hpully%d", itrkLy), Form("TRD1Dy pully-tht; dy/dz_{trk}; pull Y"), 100, -0.5, 0.5, 100, -5, 5);
      modDir->Add(fModView[itrkLy][0].hpullYtht);
      itrkLy++;

      // ===============================================================================
      // ToF Sm0 & Sm3 (Typ0) - station 5
      modDir = fOutFolder.AddFolder("ToF", "ToF reziduals to CA");
      smDir  = modDir->AddFolder("SM0_0", "ToF reziduals to CA for Sm0 and Sm3 type 0");
      for (int irpc(0); irpc < 5; irpc++) {
        fModView[itrkLy].push_back(ModuleView());
        ModuleView* cdet = &fModView[itrkLy][irpc];
        cdet->hdXx.first = 1;
        cdet->hdXx.second =
          new TH2D(Form("hxx%d%d", itrkLy, irpc), Form("ToF Sm0/0 Rpc%d x-dx; X_{ToF} (cm); #Delta X (cm)", irpc), 900,
                   -25, 65, 60, -3, 3);
        smDir->Add(cdet->hdXx.second);
        cdet->hdYy.first = 1;
        cdet->hdYy.second =
          new TH2D(Form("hyy%d%d", itrkLy, irpc), Form("ToF Sm0/0 Rpc%d y-dy; Y_{ToF} (cm); #Delta Y (cm)", irpc), 160,
                   -80, 80, 60, -3, 3);
        smDir->Add(cdet->hdYy.second);
        cdet->hdT.first = 1.;
        cdet->hdT.second =
          new TH2D(Form("ht%d%d", itrkLy, irpc), Form("ToF Sm0/0 Rpc%d x-dt; X_{ToF} (cm); #Delta T (ns)", irpc), 900,
                   -25, 65, 70, -20, 50);
        smDir->Add(cdet->hdT.second);
        cdet->hpullXphi =
          new TH2D(Form("hpullx%d%d", itrkLy, irpc), Form("ToF Sm0/0 Rpc%d pullx-phi; dx/dz_{ToF}; pull X", irpc), 160,
                   -0.3, 0.5, 100, -5, 5);
        smDir->Add(cdet->hpullXphi);
        cdet->hpullYtht =
          new TH2D(Form("hpully%d%d", itrkLy, irpc), Form("ToF Sm0/0 Rpc%d pully-tht; dy/dz_{ToF}; pull Y", irpc), 180,
                   -0.45, 0.45, 100, -5, 5);
        smDir->Add(cdet->hpullYtht);
      }
      itrkLy++;
      // ToF Sm1 (Typ0) - station 6
      smDir = modDir->AddFolder("SM1_0", "ToF reziduals to CA for Sm1 type 0");
      for (int irpc(0); irpc < 5; irpc++) {
        fModView[itrkLy].push_back(ModuleView());
        ModuleView* cdet = &fModView[itrkLy][irpc];
        cdet->hdXx.first = 1;
        cdet->hdXx.second =
          new TH2D(Form("hxx%d%d", itrkLy, irpc), Form("ToF Sm1/0 Rpc%d x-dx; X_{ToF} (cm); #Delta X (cm)", irpc), 500,
                   -25, 25, 60, -3, 3);
        smDir->Add(cdet->hdXx.second);
        cdet->hdYy.first = 1;
        cdet->hdYy.second =
          new TH2D(Form("hyy%d%d", itrkLy, irpc), Form("ToF Sm1/0 Rpc%d y-dy; Y_{ToF} (cm); #Delta Y (cm)", irpc), 160,
                   -80, 80, 60, -3, 3);
        smDir->Add(cdet->hdYy.second);
        cdet->hdT.first = 1.;
        cdet->hdT.second =
          new TH2D(Form("ht%d%d", itrkLy, irpc), Form("ToF Sm1/0 Rpc%d x-dt; X_{ToF} (cm); #Delta T (ns)", irpc), 500,
                   -25, 25, 70, -20, 50);
        smDir->Add(cdet->hdT.second);
        cdet->hpullXphi =
          new TH2D(Form("hpullx%d%d", itrkLy, irpc), Form("ToF Sm1/0 Rpc%d pullx-phi; dx/dz_{ToF}; pull X", irpc), 120,
                   -0.3, 0.3, 100, -5, 5);
        smDir->Add(cdet->hpullXphi);
        cdet->hpullYtht =
          new TH2D(Form("hpully%d%d", itrkLy, irpc), Form("ToF Sm1/0 Rpc%d pully-tht; dy/dz_{ToF}; pull Y", irpc), 180,
                   -0.45, 0.45, 100, -5, 5);
        smDir->Add(cdet->hpullYtht);
      }
      itrkLy++;
      // ToF Sm0 (Typ2) - station 7
      smDir = modDir->AddFolder("SM0_2", "ToF reziduals to CA for Sm0 type 2");
      for (int irpc(0); irpc < 5; irpc++) {
        fModView[itrkLy].push_back(ModuleView());
        ModuleView* cdet = &fModView[itrkLy][irpc];
        cdet->hdXx.first = 1;
        cdet->hdXx.second =
          new TH2D(Form("hxx%d%d", itrkLy, irpc), Form("ToF Sm0/2 Rpc%d x-dx; X_{ToF} (cm); #Delta X (cm)", irpc), 500,
                   -50, 50, 60, -3, 3);
        smDir->Add(cdet->hdXx.second);
        cdet->hdYy.first = 1;
        cdet->hdYy.second =
          new TH2D(Form("hyy%d%d", itrkLy, irpc), Form("ToF Sm0/2 Rpc%d y-dy; Y_{ToF} (cm); #Delta Y (cm)", irpc), 160,
                   -80, 80, 60, -3, 3);
        smDir->Add(cdet->hdYy.second);
        cdet->hdT.first = 1.;
        cdet->hdT.second =
          new TH2D(Form("ht%d%d", itrkLy, irpc), Form("ToF Sm0/2 Rpc%d x-dt; X_{ToF} (cm); #Delta T (ps)", irpc), 500,
                   -50, 50, 100, -500, 500);
        smDir->Add(cdet->hdT.second);
        cdet->hpullXphi =
          new TH2D(Form("hpullx%d%d", itrkLy, irpc), Form("ToF Sm0/2 Rpc%d pullx-phi; dx/dz_{ToF}; pull X", irpc), 120,
                   -0.3, 0.3, 100, -5, 5);
        smDir->Add(cdet->hpullXphi);
        cdet->hpullYtht =
          new TH2D(Form("hpully%d%d", itrkLy, irpc), Form("ToF Sm0/2 Rpc%d pully-tht; dy/dz_{ToF}; pull Y", irpc), 180,
                   -0.45, 0.45, 100, -5, 5);
        smDir->Add(cdet->hpullYtht);
      }
      itrkLy++;
      // ToF Sm2 (Typ0) - station 8
      smDir = modDir->AddFolder("SM2_0", "ToF reziduals to CA for Sm2 type 0");
      for (int irpc(0); irpc < 5; irpc++) {
        fModView[itrkLy].push_back(ModuleView());
        ModuleView* cdet = &fModView[itrkLy][irpc];
        cdet->hdXx.first = 1;
        cdet->hdXx.second =
          new TH2D(Form("hxx%d%d", itrkLy, irpc), Form("ToF Sm2/0 Rpc%d x-dx; X_{ToF} (cm); #Delta X (cm)", irpc), 600,
                   -30, 30, 60, -3, 3);
        smDir->Add(cdet->hdXx.second);
        cdet->hdYy.first = 1;
        cdet->hdYy.second =
          new TH2D(Form("hyy%d%d", itrkLy, irpc), Form("ToF Sm2/0 Rpc%d y-dy; Y_{ToF} (cm); #Delta Y (cm)", irpc), 160,
                   -80, 80, 60, -3, 3);
        smDir->Add(cdet->hdYy.second);
        cdet->hdT.first = 1.;
        cdet->hdT.second =
          new TH2D(Form("ht%d%d", itrkLy, irpc), Form("ToF Sm2/0 Rpc%d x-dt; X_{ToF} (cm); #Delta T (ps)", irpc), 600,
                   -30, 30, 100, -500, 500);
        smDir->Add(cdet->hdT.second);
        cdet->hpullXphi =
          new TH2D(Form("hpullx%d%d", itrkLy, irpc), Form("ToF Sm2/0 Rpc%d pullx-phi; dx/dz_{ToF}; pull X", irpc), 120,
                   -0.3, 0.3, 100, -5, 5);
        smDir->Add(cdet->hpullXphi);
        cdet->hpullYtht =
          new TH2D(Form("hpully%d%d", itrkLy, irpc), Form("ToF Sm2/0 Rpc%d pully-tht; dy/dz_{ToF}; pull Y", irpc), 180,
                   -0.45, 0.45, 100, -5, 5);
        smDir->Add(cdet->hpullYtht);
      }
      itrkLy++;

      // ===============================================================================
      // TRG - upstream projections
      modDir = fOutFolder.AddFolder("Target", "Target tomography with CA");
      iprjLy = itrkLy;
      for (auto zprj : fProjView) {
        LOG(debug) << GetName() << "::Init: Project trks @ " << zprj.first;
        fModView[iprjLy].push_back(ModuleView());
        fModView[iprjLy][0].hdXx.first = 1;
        fModView[iprjLy][0].hdXx.second =
          new TH2D(Form("hxy%d", iprjLy - itrkLy), Form("Project @ z = %+5.2f; X_{trk} (cm); Y_{trk} (cm)", zprj.first),
                   700, -25, 10, 100, -5, 5);
        modDir->Add(fModView[iprjLy][0].hdXx.second);
        fModView[iprjLy][0].hdYy.first = 1;
        fModView[iprjLy][0].hdYy.second =
          new TH2D(Form("hxt%d", iprjLy), Form("Project @ z = %+5.2f; X_{trk} (cm); T_{trk} - T_{ev} (ns)", zprj.first),
                   700, -25, 10, 350, -15.5, 5.5);
        modDir->Add(fModView[iprjLy][0].hdYy.second);
        iprjLy++;
      }
      break;
    case kMcbm24: LOG(info) << GetName() << "::Init: Setup config for \"mCBM 2024\". Not implemented."; break;
    case kDefault: LOG(info) << GetName() << "::Init: Setup config for \"DEFAULT\". Not implemented."; break;
  }

  return kSUCCESS;
}

//_____________________________________________________________________
void CbmRecoQaTask::Exec(Option_t*)
{
  LOG(debug) << GetName() << "::Exec : Evs[" << (fEvents ? fEvents->GetEntriesFast() : 0) << "] Trks["
             << fTracks->GetEntriesFast() << "]";
  int itrack(0), nnodes(0);
  for (auto trackObj : *fTracks) {
    //printf("GlobalTrack %d\n", itrack);
    auto track = dynamic_cast<const CbmGlobalTrack*>(trackObj);
    CbmKfTrackFitter::Track trkKf;
    if (!fFitter.CreateGlobalTrack(trkKf, *track)) {
      LOG(fatal) << GetName() << "::Exec: can not create the track for the fit! ";
      break;
    }
    trkKf.fMsQp0      = 1. / 0.5;
    trkKf.fIsMsQp0Set = true;
    if (!nnodes) nnodes = (int) trkKf.fNodes.size();

    // minimalistic QA tool for tracks used for target projection
    int nhit[(size_t) ECbmModuleId::kLastModule] = {0};
    for (auto& n : trkKf.fNodes) {
      if (n.fHitSystemId == ECbmModuleId::kNotExist) continue;
      n.fIsTimeSet = false;
      n.fIsXySet   = false;
      trkKf.MakeConsistent();
      fFitter.FitTrack(trkKf);
      double dx = n.fMxy.X()[0] - n.fTrack.X()[0], dy = n.fMxy.Y()[0] - n.fTrack.Y()[0],
             dt    = n.fMt.T()[0] - n.fTrack.Time()[0],
             pullx = dx / sqrt(n.fMxy.Dx2()[0] + n.fTrack.GetCovariance(0, 0)[0]),
             pully = dy / sqrt(n.fMxy.Dy2()[0] + n.fTrack.GetCovariance(1, 1)[0]);
      // pullt = dt / sqrt(n.fMt.Dt2()[0] + n.fTrack.GetCovariance(5, 5)[0]);

      // add granularity for specific systems
      int modId = 0;
      switch (n.fHitSystemId) {
        case ECbmModuleId::kTof:
          modId = CbmTofAddress::GetRpcId(n.fHitAddress);
          nhit[(int) n.fHitSystemId]++;
          break;
        default: nhit[(int) n.fHitSystemId]++; break;
      }

      // check that we have correctly created all views
      if (fModView.find(n.fMaterialLayer) == fModView.end()) {
        LOG(debug) << GetName() << "::Exec: view for tracking layer " << n.fMaterialLayer << " not defined.";
        continue;
      }
      if (fModView[n.fMaterialLayer].size() <= (size_t) modId) {
        LOG(debug) << GetName() << "::Exec: view for tracking layer " << n.fMaterialLayer << " @ " << modId
                   << " not defined.";
        continue;
      }
      ModuleView* mod = &fModView[n.fMaterialLayer][modId];
      if (!mod) {
        LOG(warn) << GetName() << "::Exec: can not access view for tracking layer " << n.fMaterialLayer << " @ "
                  << modId;
        continue;
      }
      if (mod->hdXx.second) mod->hdXx.second->Fill(n.fMxy.X()[0], mod->hdXx.first * dx);
      if (mod->hdYy.second) mod->hdYy.second->Fill(n.fMxy.Y()[0], mod->hdYy.first * dy);
      if (mod->hdT.second) {
        if (n.fHitSystemId == ECbmModuleId::kTrd
            && CbmTrdAddress::GetModuleAddress(n.fHitAddress)
                 == 37) {  // TODO mark generically the low resolution coordinate of TRD
          mod->hdT.second->Fill(n.fMxy.Y()[0], mod->hdT.first * dt);
        }
        else
          mod->hdT.second->Fill(n.fMxy.X()[0], mod->hdT.first * dt);
      }
      if (mod->hpullXphi) mod->hpullXphi->Fill(n.fTrack.Tx()[0], pullx);
      if (mod->hpullYtht) mod->hpullYtht->Fill(n.fTrack.Ty()[0], pully);
      n.fIsTimeSet = true;
      n.fIsXySet   = true;
    }

    // track QA filtering
    if (!FilterTrack(nhit)) continue;

    // Fit track with all hits ON for vx projection
    int iprj(0);
    for (auto zprj : fProjView) {
      CbmKfTrackFitter::FitNode n = CbmKfTrackFitter::FitNode();
      n.fZ                        = zprj.first;
      n.fIsXySet                  = true;
      n.fReference1               = iprj++;
      trkKf.fNodes.push_back(n);
    }
    trkKf.MakeConsistent();
    fFitter.FitTrack(trkKf);
    for (auto& n : trkKf.fNodes) {
      if (n.fReference1 < 0) continue;
      if (fModView.find(n.fReference1 + nnodes) == fModView.end()) {
        LOG(debug) << GetName() << "::Exec: view for tracking layer " << n.fReference1 + nnodes << " not defined.";
        continue;
      }
      ModuleView* mod = &fModView[n.fReference1 + nnodes][0];

      if (mod->hdXx.second)
        mod->hdXx.second->Fill(mod->hdXx.first * n.fTrack.X()[0], mod->hdXx.first * n.fTrack.Y()[0]);
      if (mod->hdYy.second) mod->hdYy.second->Fill(mod->hdYy.first * n.fTrack.X()[0], n.fTrack.Time()[0]);
    }
    itrack++;
  }
}

//_____________________________________________________________________
void CbmRecoQaTask::SetProjections(std::vector<double> vzpj)
{
  fProjView.clear();
  for (auto zprj : vzpj) {
    fProjView[zprj] = new TH2D(Form("hxx"), Form("; X (cm); Y (cm)"), 450, -25, 20, 450, -25, 20);
  }
}

//_____________________________________________________________________
bool CbmRecoQaTask::FilterTrack(int* ptr)
{
  bool ret = false;
  switch (fSetupClass) {
    case kMcbm22: {
      int* nhit = ptr;
      if (nhit[int(ECbmModuleId::kSts)] == 2 && nhit[int(ECbmModuleId::kTrd2d)] && nhit[int(ECbmModuleId::kTof)])
        ret = true;
      break;
    }
    case kMcbm24: break;
    case kDefault: break;
  }
  return ret;
}

//_____________________________________________________________________
void CbmRecoQaTask::Finish()
{
  FairSink* sink = FairRootManager::Instance()->GetSink();
  sink->WriteObject(&fOutFolder, nullptr);
}
